import functools
import logging
import math
import re
import urllib.error
from pathlib import Path
from urllib.request import urlopen

import cv2
import numpy as np
from colormath.color_conversions import convert_color
from colormath.color_diff import delta_e_cie2000
from colormath.color_objects import sRGBColor, LabColor

from stone.utils import is_url, extract_filename_and_extension, alphabet_id, ArgumentError

LOG = logging.getLogger(__name__)

DEFAULT_TONE_PALETTE = {
    # Default skin tone palette
    "perla": [
        "#373028",
        "#422811",
        "#513b2e",
        "#6f503c",
        "#81654f",
        "#9d7a54",
        "#bea07e",
        "#e5c8a6",
        "#e7c1b8",
        "#f3dad6",
        "#fbf2f3",
    ],
    # Refer to this paper:
    # Ostfeld, M. C., & Yadon, N. (2022). Skin color, power, and politics in America. Russell Sage Foundation.
    "yadon-ostfeld": [
        "#36251d",
        "#48352c",
        "#614539",
        "#755848",
        "#886958",
        "#9b7966",
        "#b18972",
        "#c29c88",
        "#d4afa3",
        "#e6c6bf",
    ],
    # Refer to this paper:
    # Proyecto sobre discriminación étnico-racial en México (PRODER). El Colegio de México. https://discriminacion.colmex.mx/encuesta-proder/
    "proder": [
        "#654d3e",
        "#775741",
        "#876249",
        "#946c51",
        "#a0765a",
        "#a87f64",
        "#b1886c",
        "#b69279",
        "#be9d86",
        "#c5a691",
        "#c8ac99",
    ],
    # Refer to this paper:
    # Leigh, A., & Susilo, T. (2009). Is voting skin-deep? Estimating the effect of candidate ballot photographs on election outcomes.
    # Journal of Economic Psychology, 30(1), 61-70.
    "bw": [
        "#FFFFFF",
        "#F0F0F0",
        "#E0E0E0",
        "#D0D0D0",
        "#C0C0C0",
        "#B0B0B0",
        "#A0A0A0",
        "#909090",
        "#808080",
        "#707070",
        "#606060",
        "#505050",
        "#404040",
        "#303030",
        "#202020",
        "#101010",
        "#000000",
    ],
}

def default_tone_labels(tone_palette, prefix:str=""):
    prefix = prefix or ""
    return [f"{prefix}{alphabet_id(i)}" for i in range(len(tone_palette))]

@functools.lru_cache(maxsize=128)  # Python 3.2+
def normalize_color(color):
    hex_color_pattern = re.compile(r"^#([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})$")
    decimal_color_pattern = re.compile(
        r"^(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)"
        r",\s*(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)"
        r",\s*(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$"
    )
    if decimal_color_pattern.match(color):
        r, g, b = map(int, color.split(","))
        color = "#{:02X}{:02X}{:02X}".format(r, g, b)
        return color
    if hex_color_pattern.match(color):
        return color.upper()
    raise ArgumentError(f"Invalid color code: {color}")


# @functools.lru_cache(maxsize=128)  # Python 3.2+
def normalize_palette(palette):
    return [normalize_color(color) for color in palette]


def load_image(filename_or_url, flags=cv2.IMREAD_COLOR):
    if isinstance(filename_or_url, str):
        if is_url(filename_or_url):
            base_filename, extension = extract_filename_and_extension(filename_or_url)
            image = image_from_url(filename_or_url, flags)
            return image, base_filename, extension
        filename_or_url = Path(filename_or_url)
    if not Path(filename_or_url).exists():
        raise FileNotFoundError(f"{filename_or_url} is not found.")
    base_filename, extension = filename_or_url.stem, filename_or_url.suffix
    filename = str(filename_or_url.resolve())
    image = cv2.imread(filename, flags)
    return image, base_filename, extension


def image_from_url(url, flags=cv2.IMREAD_COLOR):
    """
    Read image from url.
    Refer to https://stackoverflow.com/a/55026951/8860079
    :param url:
    :param flags:
    :return:
    """
    try:
        resp = urlopen(url)
        image = np.asarray(bytearray(resp.read()), dtype="uint8")
        image = cv2.imdecode(image, flags)
    except urllib.error.HTTPError as e:
        raise FileNotFoundError(f"{url} is not found.") from e
    except Exception as e:
        raise ArgumentError(f"{url} is not a valid image.") from e
    return image


def create_color_bar(height, width, color):
    bar = np.zeros((height, width, 3), np.uint8)
    bar[:] = color
    return bar


def is_black_white(image, threshold=192) -> bool:
    """
    Check if the image is black and white
    :param image:
    :param threshold:
    :return:
    """
    # Reading Images
    if len(image.shape) == 2:
        return True
    h, w, *_ = image.shape

    # Extracting Standard Deviation
    std = np.std(image, axis=2)
    below_t = np.sum(np.where(std <= 25))
    prob_bt = below_t / (h * w)

    return prob_bt >= threshold


def resize(image, width: int = -1, height: int = -1):
    """
    Resize the image, -1 means auto, but the image won't be resized if both width and height are -1
    :param image:
    :param width: -1 means auto
    :param height: -1 means auto
    :return:
    """
    if width < 0 and height < 0:
        return image
    elif width < 0:
        ratio = height / image.shape[0]
        width = int(image.shape[1] * ratio)
    elif height < 0:
        ratio = width / image.shape[1]
        height = int(image.shape[0] * ratio)

    return cv2.resize(image, (width, height))


def detect_faces(
    image,
    scaleFactor=1.1,
    minNeighbors=5,
    minSize=(30, 30),
    biggest_only=True,
    is_bw=False,
    threshold=0.3,
):
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    gray = cv2.equalizeHist(gray)

    cascade = cv2.CascadeClassifier(cv2.data.haarcascades + "haarcascade_frontalface_default.xml")

    flags = (cv2.CASCADE_SCALE_IMAGE | cv2.CASCADE_FIND_BIGGEST_OBJECT) if biggest_only else cv2.CASCADE_SCALE_IMAGE
    faces = cascade.detectMultiScale(
        gray,
        scaleFactor=scaleFactor,
        minNeighbors=minNeighbors,
        minSize=minSize,
        flags=flags,
    )
    if len(faces) == 0:
        return []
    # Change the format of faces from (x, y, w, h) to (x, y, x+w, y+h)
    faces[:, 2:] += faces[:, :2]
    return [face for face in faces if is_face(face, image, is_bw, threshold)]


def is_face(face_coord, image, is_bw, threshold=0.3):
    """
    Check if the face is a real face.
    Method: detect the skin area in the "face" and check if the skin area is larger than the threshold
    :param face_coord:
    :param image:
    :param is_bw:
    :param threshold:
    :return:
    """
    x1, y1, x2, y2 = face_coord
    face_image = image[y1:y2, x1:x2]
    detect_skin_fn = detect_skin_in_bw if is_bw else detect_skin_in_color
    _, skin_mask = detect_skin_fn(face_image)
    skin_pixels = cv2.countNonZero(skin_mask)
    total_pixels = face_image.shape[0] * face_image.shape[1]
    skin_ratio = skin_pixels / total_pixels
    return skin_ratio >= threshold


def mask_face(image, face):
    x1, y1, x2, y2 = face
    mask = np.zeros(image.shape[:2], dtype=np.uint8)
    mask[y1:y2, x1:x2] = 255  # Fill with white color
    image = cv2.bitwise_and(image, image, mask=mask)
    return image


def detect_skin_in_bw(image):
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    _, threshold = cv2.threshold(gray, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (5, 5))
    skin_mask = cv2.morphologyEx(threshold, cv2.MORPH_CLOSE, kernel)

    skin = cv2.bitwise_and(image, image, mask=skin_mask)
    all_0 = np.isclose(skin, 0).all()
    return image if all_0 else skin, skin_mask


def detect_skin_in_color(image):
    # Converting from BGR Colors Space to HSV
    img = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)

    # Defining skin Thresholds
    low_hsv = np.array([0, 48, 80], dtype=np.uint8)
    high_hsv = np.array([20, 255, 255], dtype=np.uint8)

    skin_mask = cv2.inRange(img, low_hsv, high_hsv)
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3))
    skin_mask = cv2.morphologyEx(skin_mask, cv2.MORPH_OPEN, kernel)
    skin_mask = cv2.morphologyEx(skin_mask, cv2.MORPH_CLOSE, kernel)
    skin_mask = cv2.GaussianBlur(skin_mask, ksize=(3, 3), sigmaX=0)

    skin = cv2.bitwise_and(image, image, mask=skin_mask)

    all_0 = np.isclose(skin, 0).all()
    return image if all_0 else skin, skin_mask


def draw_rects(image, *rects, color=(255, 0, 0), thickness=2):
    for x1, y1, x2, y2 in rects:
        cv2.rectangle(image, (x1 - 1, y1 - 1), (x2 + 1, y2 + 1), color, thickness)
    return image


def dominant_colors(image, to_bw, n_clusters=2):
    if to_bw:
        data = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        data = cv2.cvtColor(data, cv2.COLOR_GRAY2BGR)
    else:
        data = image
    data = np.reshape(data, (-1, 3))
    data = data[np.all(data != 0, axis=1)]
    data = np.float32(data)

    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 10, 1.0)
    flags = cv2.KMEANS_RANDOM_CENTERS
    compactness, labels, colors = cv2.kmeans(data, n_clusters, None, criteria, 10, flags)
    labels, counts = np.unique(labels, return_counts=True)

    order = (-counts).argsort()
    colors = colors[order]
    counts = counts[order]

    percents = counts / counts.sum()

    return colors, percents


def blur(image, degree=25):
    """
    Blur the image
    :param image:
    :param degree: The degree of blur. The bigger, the more blur
    :return:
    """
    ksize = degree, degree
    return cv2.blur(image, ksize)


def skin_tone(colors, percents, skin_tone_palette, tone_labels):
    lab_tones = [convert_color(sRGBColor.new_from_rgb_hex(rgb), LabColor) for rgb in skin_tone_palette]
    lab_colors = [convert_color(sRGBColor(rgb_r=r, rgb_g=g, rgb_b=b, is_upscaled=True), LabColor) for b, g, r in colors]
    distances = [np.sum([delta_e_cie2000(c, label) * p for c, p in zip(lab_colors, percents)]) for label in lab_tones]
    tone_id = np.argmin(distances)
    distance: float = distances[tone_id]
    tone_hex = skin_tone_palette[tone_id].upper()
    tone_label = tone_labels[tone_id]
    return tone_id, tone_hex, tone_label, distance


def classify(
    image,
    is_bw,
    to_bw,
    skin_tone_palette,
    tone_labels,
    n_dominant_colors=2,
    verbose=False,
    report_image=None,
    use_face=True,
):
    """
    Classify the skin tone of the image
    :param image: Entire image or image with non-face areas masked
    :param is_bw: Whether the image is black and white
    :param to_bw: Whether to convert the image to black and white
    :param skin_tone_palette:
    :param tone_labels:
    :param n_dominant_colors:
    :param verbose: Whether to output the report image
    :param report_image: The image to draw the report on
    :param use_face: whether to use face area for detection
    :return:
    """
    detect_skin_fn = detect_skin_in_bw if is_bw else detect_skin_in_color
    skin, skin_mask = detect_skin_fn(image)
    dmnt_colors, dmnt_pcts = dominant_colors(skin, to_bw, n_dominant_colors)
    # Generate readable strings
    hex_colors = ["#%02X%02X%02X" % tuple(np.around([r, g, b]).astype(int)) for b, g, r in dmnt_colors]
    pct_strs = ["%.2f" % p for p in dmnt_pcts]
    result = {"dominant_colors": [{"color": color, "percent": pct} for color, pct in zip(hex_colors, pct_strs)]}
    # Calculate skin tone
    tone_id, tone_hex, tone_label, distance = skin_tone(dmnt_colors, dmnt_pcts, skin_tone_palette, tone_labels)
    accuracy = round(100 - distance, 2)
    result["skin_tone"] = tone_hex
    result["tone_label"] = tone_label
    result["accuracy"] = accuracy
    if not verbose:
        return result, None

    # 0. Create initial report image
    report_image = initial_report_image(image, report_image, skin_mask, use_face, to_bw)
    bar_width = 100

    # 1. Create color bar for dominant colors
    color_bars = create_dominant_color_bar(report_image, dmnt_colors, dmnt_pcts, bar_width)

    # 2. Create color bar for a skin tone list
    palette_bars = create_tone_palette_bar(report_image, tone_id, skin_tone_palette, bar_width)

    # 3. Combine all bars and report image
    report_image = np.hstack([report_image, color_bars, palette_bars])
    msg_bar = create_message_bar(dmnt_colors, dmnt_pcts, tone_hex, distance, report_image.shape[1])
    report_image = np.vstack([report_image, msg_bar])
    return result, report_image


def initial_report_image(face_image, report_image, skin_mask, use_face, to_bw):
    report_image = face_image if report_image is None else report_image
    if to_bw:
        report_image = cv2.cvtColor(report_image, cv2.COLOR_BGR2GRAY)
        report_image = cv2.cvtColor(report_image, cv2.COLOR_GRAY2BGR)
    if use_face:
        gray = cv2.cvtColor(face_image, cv2.COLOR_BGR2GRAY)
        skin_mask = cv2.threshold(gray, 1, 255, cv2.THRESH_BINARY)[1]
    blurred_image = blur(report_image)
    non_skin_mask = cv2.bitwise_not(skin_mask)
    edges = cv2.Canny(skin_mask, 50, 150)
    report_image = cv2.bitwise_and(report_image, report_image, mask=skin_mask) + cv2.bitwise_and(
        blurred_image, blurred_image, mask=non_skin_mask
    )
    contours, _ = cv2.findContours(edges, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    cv2.drawContours(report_image, contours, -1, (255, 0, 0), 2)
    return report_image


def create_dominant_color_bar(report_image, dmnt_colors, dmnt_pcts, bar_width):
    color_bars = []
    total_height = 0
    for color, pct in zip(dmnt_colors, dmnt_pcts):
        bar_height = int(math.floor(report_image.shape[0] * pct))
        total_height += bar_height
        bar = create_color_bar(bar_height, bar_width, color)
        color_bars.append(bar)
    padding_height = report_image.shape[0] - total_height
    if padding_height > 0:
        padding = create_color_bar(padding_height, bar_width, (255, 255, 255))
        color_bars.append(padding)
    return np.vstack(color_bars)


def hex_to_bgr(hex_color):
    hex_value = hex_color.lstrip("#")
    r, g, b = [int(hex_value[i : i + 2], 16) for i in (0, 2, 4)]
    return [b, g, r]


def rgb_to_hex(rgb_color):
    r, g, b = rgb_color
    return "#%02X%02X%02X" % (r, g, b)


def create_tone_palette_bar(report_image, tone_id, skin_tone_palette, bar_width):
    palette_bars = []
    tone_height = report_image.shape[0] // len(skin_tone_palette)
    tone_bgrs = []
    for tone in skin_tone_palette:
        hex_value = tone.lstrip("#")
        r, g, b = [int(hex_value[i : i + 2], 16) for i in (0, 2, 4)]
        tone_bgrs.append([b, g, r])
        bar = create_color_bar(tone_height, bar_width, [b, g, r])
        palette_bars.append(bar)
    padding_height = report_image.shape[0] - tone_height * len(skin_tone_palette)
    if padding_height > 0:
        padding = create_color_bar(padding_height, bar_width, (255, 255, 255))
        palette_bars.append(padding)
    bar = np.vstack(palette_bars)

    padding = 1
    start_point = (padding, tone_id * tone_height + padding)
    end_point = (bar_width - padding, (tone_id + 1) * tone_height)
    bar = cv2.rectangle(bar, start_point, end_point, (255, 0, 0), 2)
    return bar


def create_message_bar(dmnt_colors, dmnt_pcts, tone_hex, distance, bar_width):
    msg_bar = create_color_bar(height=50, width=bar_width, color=(243, 239, 214))
    b, g, r = np.around(dmnt_colors[0]).astype(int)
    dominant_color_hex = "#%02X%02X%02X" % (r, g, b)
    pct = f"{dmnt_pcts[0] * 100:.2f}%"

    font, font_scale, txt_colr, thickness, line_type = (
        cv2.FONT_HERSHEY_SIMPLEX,
        0.5,
        (0, 0, 0),
        1,
        cv2.LINE_AA,
    )
    x, y = 2, 15
    msg = f"- Dominant color: {dominant_color_hex}, percent: {pct}"
    cv2.putText(msg_bar, msg, (x, y), font, font_scale, txt_colr, thickness, line_type)

    text_size, _ = cv2.getTextSize(msg, font, font_scale, thickness)
    line_height = text_size[1] + 10
    accuracy = round(100 - distance, 2)
    cv2.putText(
        msg_bar,
        f"- Skin tone: {tone_hex}, accuracy: {accuracy}",
        (x, y + line_height),
        font,
        font_scale,
        txt_colr,
        thickness,
        cv2.LINE_AA,
    )

    return msg_bar


def process_image(
    image: np.ndarray,
    is_bw: bool,
    to_bw: bool,
    skin_tone_palette: list,
    tone_labels: list = None,
    new_width=-1,
    n_dominant_colors=2,
    scaleFactor=1.1,
    minNeighbors=5,
    minSize=(30, 30),
    biggest_only=True,
    threshold=0.3,
    verbose=False,
):
    image = resize(image, new_width)

    records, report_images = [], {}
    face_coords = detect_faces(image, scaleFactor, minNeighbors, minSize, biggest_only, is_bw, threshold)
    n_faces = len(face_coords)

    if n_faces == 0:
        # If no face is detected, find skin area in the whole image and classify.
        record, report_image = classify(
            image,
            is_bw,
            to_bw,
            skin_tone_palette,
            tone_labels,
            n_dominant_colors,
            verbose=verbose,
            use_face=False,
        )
        record["face_id"] = "NA"
        records.append(record)
        report_images["NA"] = report_image
    # Otherwise, detect skin tone for each face
    for idx, face_coord in enumerate(face_coords):
        face_image = mask_face(image, face_coord)
        record, report_image = classify(
            face_image,
            is_bw,
            to_bw,
            skin_tone_palette,
            tone_labels,
            n_dominant_colors,
            verbose=verbose,
            report_image=image,
            use_face=True,
        )
        record["face_id"] = idx + 1
        records.append(record)
        report_image = face_report_image(face_coord, idx, report_image)
        report_images[idx + 1] = report_image

    return records, report_images


def show(image, title=None):
    title = f" - {title}" if title else ""
    cv2.imshow(f"Skin Tone Classifier{title}", image)
    cv2.waitKey(0)
    cv2.destroyAllWindows()


def face_report_image(face, idx, image):
    if image is None:
        return None
    x1, y1, x2, y2 = face
    width = x2 - x1
    height = 20
    bar = np.ones((height, width, 3), dtype=np.uint8) * (255, 0, 0)
    report_image = image.copy()
    report_image[y2 : y2 + height, x1:x2] = bar
    txt = f"Face {idx + 1}"
    text_color = (255, 255, 255)
    font_scale = 0.5
    thickness = 1
    text_size, _ = cv2.getTextSize(txt, cv2.FONT_HERSHEY_SIMPLEX, font_scale, thickness)
    text_x = x1 + (width - text_size[0]) // 2
    text_y = y2 + 15
    cv2.putText(
        report_image,
        txt,
        (text_x, text_y),
        cv2.FONT_HERSHEY_SIMPLEX,
        font_scale,
        text_color,
        thickness,
    )
    return report_image



import unittest
from pathlib import Path

from stone.image import default_tone_labels
from stone.utils import build_image_paths, resolve_labels, alphabet_id


class TestUtils(unittest.TestCase):
    def setUp(self):
        self.image_path = str(Path("./mock_data/images").resolve())
        # Sorted image paths
        self.expected_recursive_image_paths = [
            f"{self.image_path}/fake_img_1.gif",  # In default, sorted by the trailing number
            f"{self.image_path}/fake_img_2.jpeg",
            f"{self.image_path}/subfolder/sub_fake_img_3.gif",
            f"{self.image_path}/subfolder/sub_fake_img_4.jpeg",
            f"{self.image_path}/fake_img_10.webp",  # Sorted by length of the filename if the trailing number is the same
            f"{self.image_path}/subfolder/sub_fake_img_10.jpg",
            f"{self.image_path}/subfolder/sub_fake_img_21.png",
            f"{self.image_path}/fake_img_22.png",
            f"{self.image_path}/fake_img_100.jpg",
            f"{self.image_path}/subfolder/sub_fake_img_101.webp",
        ]

        self.expected_non_recursive_image_paths = [
            p for p in self.expected_recursive_image_paths if "subfolder" not in p
        ]

    def should_exclude_folder(self, paths, excluded_folders):
        """
        Check if the paths do not contain any of the excluded folders.
        :param paths:
        :param excluded_folders:
        :return:
        """
        self.assertTrue(
            all(
                [
                    excluded_folder != path.relative_to(self.image_path).parts[0]
                    for path in paths
                    for excluded_folder in excluded_folders
                ]
            )
        )

    def test_single_directory_recursive(self):
        image_paths = build_image_paths(self.image_path, recursive=True)
        self.assertTrue(isinstance(image_paths, list))
        self.assertEqual(len(image_paths), 10)
        self.should_exclude_folder(image_paths, ["debug", "log"])
        for i in range(len(image_paths)):
            actual = image_paths[i]
            expected = Path(self.expected_recursive_image_paths[i])
            self.assertTrue(actual.samefile(expected), msg=f"{i}: {actual} != {expected}")

    def test_single_directory_non_recursive(self):
        image_paths = build_image_paths(self.image_path, recursive=False)
        self.assertTrue(isinstance(image_paths, list))
        self.assertEqual(len(image_paths), 5)
        self.should_exclude_folder(image_paths, ["subfolder", "debug", "log"])
        self.assertListEqual(
            image_paths,
            [Path(p).resolve() for p in self.expected_non_recursive_image_paths],
        )

    def test_multiple_directories_recursive(self):
        paths = build_image_paths([self.image_path, f"{self.image_path}/subfolder"], recursive=True)
        self.assertTrue(isinstance(paths, list))
        self.assertEqual(len(paths), 10)

    def test_single_file(self):
        paths = build_image_paths(self.expected_recursive_image_paths[0])
        self.assertTrue(isinstance(paths, list))
        self.assertEqual(len(paths), 1)

    def test_multiple_files(self):
        paths = build_image_paths(self.expected_recursive_image_paths)
        self.assertTrue(isinstance(paths, list))
        self.assertEqual(len(paths), len(self.expected_recursive_image_paths))

    def test_single_url(self):
        paths = build_image_paths("http://example.com/image.jpg")
        self.assertTrue(isinstance(paths, list))
        self.assertEqual(len(paths), 1)

    def test_no_valid_images(self):
        with self.assertRaises(FileNotFoundError):
            build_image_paths("/path/to/nonexistent")

    def test_resolve_labels_in_digits(self):
        start, end, step = 1, 12, 1
        labels = [f"{start}-{end}"]
        expected = "1 2 3 4 5 6 7 8 9 10 11 12".split()
        actual = resolve_labels(labels)
        self.assertListEqual(actual, expected)

    def test_resolve_labels_in_alphabet(self):
        start, end, step = "a", "z", 2
        labels = [f"{start}-{end}-{step}"]
        expected = list("ACEGIKMOQSUWY")
        actual = resolve_labels(labels)
        self.assertListEqual(actual, expected)

    def test_resolve_labels_step_is_zero(self):
        start, end, step = "a", "z", 0
        labels = [f"{start}-{end}-{step}"]
        expected = list("ABCDEFGHIJKLMNOPQRSTUVWXYZ")
        actual = resolve_labels(labels)
        self.assertListEqual(actual, expected)

    def test_resolve_labels_start_less_than_end(self):
        start, end, step = 12, 1, 1
        labels = [f"{start}-{end}"]
        expected = labels
        actual = resolve_labels(labels)
        self.assertListEqual(actual, expected)

    def test_default_tone_labels(self):
        tone_palette = list("ABCDEFGHIJKLMNOPQRSTUVWXYZ")
        actual_labels = default_tone_labels(tone_palette)
        self.assertListEqual(
            actual_labels,
            list("ABCDEFGHIJKLMNOPQRSTUVWXYZ"),
        )

    def test_alphabet_id(self):
        self.assertEqual(alphabet_id(0), "A")
        self.assertEqual(alphabet_id(25), "Z")
        self.assertEqual(alphabet_id(26), "AA")
        self.assertEqual(alphabet_id(27), "AB")
        self.assertEqual(alphabet_id(51), "AZ")
        self.assertEqual(alphabet_id(52), "BA")
        self.assertEqual(alphabet_id(77), "BZ")
        self.assertEqual(alphabet_id(78), "CA")

    def test_default_tone_labels_with_more_than_26_tones(self):
        tone_palette = list("ABCDEFGHIJKLMNOPQRSTUVWXYZ") * 2
        actual_labels = default_tone_labels(tone_palette)
        self.assertListEqual(
            actual_labels,
            list("ABCDEFGHIJKLMNOPQRSTUVWXYZ") + ["AA", "AB", "AC", "AD", "AE", "AF", "AG", "AH", "AI", "AJ", "AK", "AL", "AM", "AN", "AO", "AP", "AQ", "AR", "AS", "AT", "AU", "AV", "AW", "AX", "AY", "AZ"],
        )


if __name__ == "__main__":
    unittest.main()



__version__ = "1.2.5.dev0"
__package_name__ = "skin-tone-classifier"
__app_name__ = "Skin Tone Classifier"
__description__ = "An easy-to-use library for skin tone classification"
__author__ = "Chenglong Ma"
__author_email__ = "chenglong.m@outlook.com"
__author_website__ = "https://chenglongma.com/"
__url__ = "https://chenglongma.com/SkinToneClassifier/"
__code__ = "https://github.com/ChenglongMa/SkinToneClassifier/"
__issues__ = "https://github.com/ChenglongMa/SkinToneClassifier/issues"
__license__ = "GPLv3"
__copyright__ = "2022"



import argparse
import functools
import logging
import os
import re
import string
import sys
from pathlib import Path
from typing import Union
from urllib.parse import urlparse

from stone.package import __version__, __package_name__, __description__, __app_name__

LOG = logging.getLogger(__name__)


class ArgumentError(ValueError):
    """
    Wrapper for argument error. This exception will be raised when the arguments are invalid.
    """

    pass


def Gooey(*args, **kwargs):
    """
    Dummy decorator for Gooey.
    Used in CLI mode to avoid the import error when the Gooey package is not installed.
    :param args:
    :param kwargs:
    :return:
    """

    def inner(func):
        return func

    return inner


@functools.cache
def alphabet_id(n:int) -> str:
    letters = string.ascii_uppercase
    n_letters = len(letters)
    if n < n_letters:
        return letters[n]

    prefix = ""
    while n >= n_letters:
        prefix += letters[(n // n_letters) - 1]
        n %= n_letters

    return prefix + letters[n]


def is_url(text):
    return urlparse(text).scheme in ["http", "https"]


def extract_filename_and_extension(url):
    """
    Extract base filename and extension from the url.
    :param url: URL with filename and extension, e.g., https://example.com/images/pic.jpg?param=value
    :return: Base filename and extension, e.g., pic, jpg
    """
    parsed_url = urlparse(url)
    path = parsed_url.path
    filename = path.split("/")[-1]
    basename, *extension = filename.split(".")
    return basename, f".{extension[0]}" if extension else None


def build_image_paths(images_paths, recursive=False):
    filenames, urls = [], []
    valid_images = ["*.jpg", "*.gif", "*.png", "*.jpeg", "*.webp", "*.tif"]
    excluded_folders = ["debug", "log"]
    if isinstance(images_paths, str):
        images_paths = [images_paths]

    for filename in images_paths:
        if is_url(filename):
            urls.append(filename)
            continue
        p = Path(filename)
        if p.is_dir():
            images = [p.glob(pattern) for pattern in valid_images]
            if recursive:
                subfolders = [f for f in p.glob("*/") if f.name not in excluded_folders]
                images.extend([sp.rglob(pattern) for pattern in valid_images for sp in subfolders])

            filenames.extend(images)
        elif p.is_file():
            filenames.append([p])
    paths = set([f.resolve() for fs in filenames for f in fs] + urls)
    paths = list(paths)
    if len(paths) == 0:
        raise FileNotFoundError("No valid images in the specified path.")
    # Sort paths by (first) number extracted from the filename string
    paths.sort(key=sort_file)
    return paths


def sort_file(path: Union[str, Path]):
    if isinstance(path, Path):
        basename = path.stem
    else:
        basename, *_ = extract_filename_and_extension(path)
    nums = re.findall(r"\d+", basename)
    return (int(nums[0]) if nums else float("inf")), basename


def is_windows():
    return sys.platform in ["win32", "cygwin"]


def is_debugging():
    gettrace = getattr(sys, "gettrace", None)
    return gettrace is not None and gettrace()


def build_arguments():
    try:
        from gooey import GooeyParser

        in_gui = True
    except ImportError:
        from argparse import ArgumentParser as GooeyParser

        in_gui = False

    kwargs = dict(formatter_class=argparse.RawTextHelpFormatter) if not in_gui else {}
    parser = GooeyParser(
        description=__description__,
        **kwargs,
    )
    kwargs = (
        {
            "gooey_options": {"show_border": False, "columns": 1},
        }
        if in_gui
        else {}
    )
    files = parser.add_argument_group(
        "Images to process",
        "The locations of images to process, which can be directories, files, or URLs.\n"
        "Multiple values are separated by space;\n"
        'You can mix folders, filenames and web links together, e.g., "/path/to/dir1 /path/to/pic.jpg https://example.com/pic.png".\n',
        **kwargs,
    )
    kwargs = {"gooey_options": {"visible": False}} if in_gui else {}

    files.add_argument(
        "-i",
        "--images",
        nargs="+",
        default=[os.getcwd()],
        metavar="Image Filenames",
        help="Image filename(s), Directories or URLs to process. Separated by space.",
        **kwargs,
    )
    if in_gui:
        files.add_argument(
            "--image_dirs",
            nargs="+",
            metavar="Image Directories",
            widget="DirChooser",
            # widget="MultiDirChooser", # fixme: enable this widget when issues are fixed
            gooey_options={
                "message": "Select directories to process",
                "initial_value": os.getcwd(),
                "default_path": os.getcwd(),
                "placeholder": "e.g., /path/to/dir1 /path/to/dir2",
            },
        )
    kwargs = dict(metavar="Recursive Search") if in_gui else {}
    files.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="Search images recursively in the specified directory.",
        **kwargs,
    )
    if in_gui:
        files.add_argument(
            "--image_files",
            nargs="+",
            metavar="Image Filenames",
            help="Add individual image file(s)",
            widget="MultiFileChooser",
            gooey_options={
                "wildcard": "All images|*.jpg;*.jpeg;*.png;*.bmp;*.gif;*.tif;*.webp|"
                "JPG (*.jpg)|*.jpg|"
                "JPEG (*.jpeg)|*.jpeg|"
                "PNG (*.png)|*.png|"
                "BMP (*.bmp)|*.bmp|"
                "GIF (*.gif)|*.gif|"
                "TIFF (*.tif)|*.tif|"
                "WEBP (*.webp)|*.webp|"
                "All files (*.*)|*.*",
                "message": "Select the image file(s) to process",
                "default_dir": os.getcwd(),
                "full_width": False,
                "placeholder": "e.g., a.jpg b.png",
            },
        )

        files.add_argument(
            "--image_urls",
            nargs="+",
            metavar="Image URLs",
            help="Add image URLs",
            gooey_options={
                "full_width": False,
                "placeholder": "e.g., https://example.com/a.jpg https://example.com/b.png",
            },
        )

    kwargs = {"gooey_options": {"show_border": False, "columns": 2}} if in_gui else {}
    images = parser.add_argument_group(
        "Image Settings",
        **kwargs,
    )
    bw_option = "black/white" if in_gui else "bw"
    images.add_argument(
        "-t",
        "--image_type",
        default="auto",
        metavar="Image Type",
        help="Specify whether the input image(s) is/are colored or black/white.\n"
        f'Defaults to "auto", which will be detected automatically. Other options are "color" and "{bw_option}".\n',
        choices=["auto", "color", bw_option],
    )
    kwargs = {"gooey_options": {"full_width": True}} if in_gui else {}
    images.add_argument(
        "-p",
        "--palette",
        nargs="+",
        metavar="Palette",
        help="Skin tone palette;\n"
        "Valid choices are 'perla', 'yadon-ostfeld', 'proder'.\n"
        'You can also input RGB hex values leading by "#" or RGB values separated by comma(,),\n'
        "E.g., #373028 #422811 or 255,255,255 100,100,100\n"
        "Leave blank to use the 'perla' palette.\n",
        **kwargs,
    )
    images.add_argument(
        "-l",
        "--labels",
        nargs="+",
        metavar="Labels",
        help="Skin tone labels;\n"
        "Leave blank to use the default values: the uppercase alphabet list leading by the image type ('C' for 'color'; 'B' for 'Black&White'), "
        "e.g., ['CA', 'CB', ..., 'CZ'] or ['BA', 'BB', ..., 'BZ'].\n"
        "Since v1.2.0, supports range of labels, e.g., 'A-Z' or '1-10'.\n"
        "Refer to https://github.com/ChenglongMa/SkinToneClassifier#3-specify-category-labels for more details.",
        **kwargs,
    )

    kwargs = dict(metavar="Convert to Black/White") if in_gui else {}
    images.add_argument(
        "-bw",
        "--black_white",
        action="store_true",
        help="Whether to convert the input to black/white image(s)?\n"
        "If true, the app will convert the input to black/white image(s) and use the black/white palette for classification.",
        **kwargs,
    )

    kwargs = (
        {"gooey_options": {"initial_value": 2, "min": 1, "max": 99999, "full_width": False}, "widget": "IntegerField"}
        if in_gui
        else {}
    )
    images.add_argument(
        "--n_colors",
        metavar="Number of Dominant Colors",
        type=int,
        help="Specify the number of dominant colors to be extracted.\n"
        "The colors will be used to compare with the colors in the palette.\n",
        default=2,
        **kwargs,
    )

    kwargs = (
        {
            "gooey_options": {"initial_value": 250, "min": 10, "max": 99999, "full_width": False},
            "widget": "IntegerField",
        }
        if in_gui
        else {}
    )
    images.add_argument(
        "--new_width",
        type=int,
        metavar="New Width (pixels)",
        help="Resize the images with the specified width.\n"
        "Sometimes smaller images will be processed faster and more accurately.\n"
        "No resizing will be performed if the value is negative.",
        default=250,
        **kwargs,
    )

    kwargs = {"gooey_options": {"show_border": True}} if in_gui else {}
    outputs = parser.add_argument_group("Output Settings", **kwargs)

    kwargs = (
        {
            "gooey_options": {"message": "Select the output directory", "default_path": os.getcwd()},
            "widget": "DirChooser",
        }
        if in_gui
        else {}
    )
    outputs.add_argument(
        "-o",
        "--output",
        metavar="Output Directory",
        default=os.getcwd(),
        help="Specify the path of output file, defaults to current directory.",
        **kwargs,
    )

    kwargs = dict(metavar="Generate Report Images") if in_gui else {}
    outputs.add_argument(
        "-d",
        "--debug",
        action="store_true",
        default=in_gui,
        help="Whether to generate report images?\n"
        "If true, the report images will be saved in the '<OUTPUT_DIRECTORY>/debug' directory.",
        **kwargs,
    )

    kwargs = {"gooey_options": {"show_border": False, "columns": 2}} if in_gui else {}
    advanced = parser.add_argument_group(
        "Advanced Settings",
        "For advanced users only, please refer to https://stackoverflow.com/a/20805153/8860079",
        **kwargs,
    )

    kwargs = (
        {"gooey_options": {"initial_value": 1.1, "min": 0.1, "max": 2.0}, "widget": "DecimalField"} if in_gui else {}
    )
    advanced.add_argument(
        "--scale",
        type=float,
        metavar="Scale",
        help="Specify how much the image size is reduced at each image scale.",
        default=1.1,
        **kwargs,
    )

    kwargs = {"gooey_options": {"initial_value": 5, "min": 1, "max": 99999}, "widget": "IntegerField"} if in_gui else {}
    advanced.add_argument(
        "--min_nbrs",
        type=int,
        metavar="Minimum Neighbors",
        help="Specify how many neighbors each candidate rectangle should have to retain it.\n"
        "Higher value results in less detections but with higher quality.",
        default=5,
        **kwargs,
    )
    default_min_width = 90
    default_min_height = 90

    kwargs = {"gooey_options": {"visible": False}} if in_gui else {}
    advanced.add_argument(
        "--min_size",
        type=int,
        nargs="+",
        metavar="Minimum Possible Face Size, format: <Width Height>",
        help=f'Specify the minimum possible face size. Faces smaller than that are ignored, defaults to "{default_min_width} {default_min_height}".',
        default=(default_min_width, default_min_height),
        **kwargs,
    )
    if in_gui:
        min_size = advanced.add_argument_group(
            "Minimum Possible Face Size (pixels)",
            'Specify the minimum possible face size. Faces smaller than that are ignored, defaults to "90 90".',
            gooey_options={"show_border": True, "columns": 2},
        )

        min_size.add_argument(
            "--min_width",
            type=int,
            metavar="Minimum Width",
            # help="Specify the minimum possible face width. Faces smaller than that are ignored, defaults to 90.",
            default=default_min_width,
            widget="IntegerField",
            gooey_options={"initial_value": default_min_width, "min": 10, "max": 99999},
        )

        min_size.add_argument(
            "--min_height",
            type=int,
            metavar="Minimum Height",
            # help="Specify the minimum possible face height. Faces smaller than that are ignored, defaults to 90.",
            default=default_min_height,
            widget="IntegerField",
            gooey_options={"initial_value": default_min_height, "min": 10, "max": 99999},
        )

    kwargs = (
        {"gooey_options": {"initial_value": 0.15, "min": 0.01, "max": 1.0}, "widget": "DecimalField"} if in_gui else {}
    )
    advanced.add_argument(
        "--threshold",
        type=float,
        metavar="Minimum Possible Face Proportion",
        help="Specify the minimum proportion of the skin area required to identify the face, defaults to 0.15.",
        default=0.15,
        **kwargs,
    )

    kwargs = {"gooey_options": {"initial_value": 0, "min": 0, "max": 99999}, "widget": "IntegerField"} if in_gui else {}
    advanced.add_argument(
        "--n_workers",
        type=int,
        metavar="Number of CPU Workers",
        help="Specify the number of workers to process the images.\n"
        "0 means the total number of CPU cores in the system.",
        default=0,
        **kwargs,
    )

    kwargs = dict(gooey_options={"visible": False}) if in_gui else {}
    advanced.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
        help="Show the version number and exit.",
        **kwargs,
    )
    args = parser.parse_args()
    images = args.images or []
    if getattr(args, "image_dirs", False):
        images.extend(args.image_dirs)
    if getattr(args, "image_files", False):
        images.extend(args.image_files)
    if getattr(args, "image_urls", False):
        images.extend(args.image_urls)
    args.images = images
    if (
        tuple(args.min_size) == (default_min_width, default_min_height)
        and getattr(args, "min_width", False)
        and getattr(args, "min_height", False)
    ):
        args.min_size = (args.min_width, args.min_height)
    return args


def resolve_labels(labels):
    if not labels or len(labels) != 1:
        return labels
    label = labels[0]

    separator = r"[-,~:;_]"
    pattern = rf"^([a-zA-Z0-9]+){separator}([a-zA-Z0-9]+)(?:{separator}([-+]?\d+))?$"
    match = re.match(pattern, label)
    if match is None:
        return labels
    start, end, step = match.groups()
    if not step:
        step = 1
    else:
        step = int(step)
    if step == 0:
        LOG.warning(f"The specified step in the '--label' setting ('{label}') cannot be 0; resetting to 1.")
        step = 1
    if step < 0:
        start, end = end, start

    if start.isdigit() and end.isalpha() or start.isalpha() and end.isdigit():
        LOG.warning(
            f"Invalid '--label' setting ('{label}'): The start value ({start}) and the end value ({end}) should be both digits or both letters."
        )
        return labels
    if start >= end:
        LOG.warning(
            f"Invalid '--label' setting ('{label}'): The start value ({start}) should be less than the end value ({end})."
        )
        return labels
    if start.isdigit() and end.isdigit():
        start, end = int(start), int(end)
        return [str(i) for i in range(start, end + 1, step)]
    if start.isalpha() and end.isalpha():
        start, end = start.upper(), end.upper()
        return [chr(i) for i in range(ord(start), ord(end) + 1, step)]
    return labels


def get_latest_version_from_pypi(package_name):
    try:
        import requests

        response = requests.get(f"https://pypi.org/pypi/{package_name}/json")
        response.raise_for_status()

        data = response.json()
        latest_version = data["info"]["version"]
        return latest_version
    except Exception:
        pass


def check_version():
    if "STONE_UPGRADE_FLAG" in os.environ:
        return
    try:
        from packaging.version import parse
        import importlib.metadata

        latest_version = get_latest_version_from_pypi(__package_name__)
        if not latest_version:
            return
        distribution = importlib.metadata.distribution(__package_name__)
        installed_version = distribution.version
        if parse(installed_version) < parse(latest_version):
            from colorama import just_fix_windows_console, Fore

            just_fix_windows_console()
            print(
                Fore.YELLOW + f"You are using an outdated version of {__package_name__} ({installed_version}).\n"
                f"Please upgrade to the latest version ({latest_version}) with the following command:\n",
                Fore.GREEN + f"pip install {__package_name__}[all] --upgrade\n" + Fore.RESET,
            )
            os.environ["STONE_UPGRADE_FLAG"] = "1"
    except Exception:
        pass



import numpy as np
from stone.api import process
from stone.image import DEFAULT_TONE_PALETTE, show
from stone.utils import __version__, check_version

setattr(np, "asscalar", lambda x: np.asarray(x).item())

__all__ = ["process", "DEFAULT_TONE_PALETTE", "show", "__version__"]

check_version()



import logging
from pathlib import Path
from typing import Union, Literal, List

import cv2

from stone.image import (
    load_image,
    is_black_white,
    DEFAULT_TONE_PALETTE,
    process_image,
    normalize_palette, default_tone_labels,
)
from stone.utils import ArgumentError

LOG = logging.getLogger(__name__)


def process(
    filename_or_url: Union[str, Path],
    image_type: Literal["auto", "color", "bw"] = "auto",
    tone_palette: Union[List[str], Literal["perla", "yadon-ostfeld", "proder", "bw"]] = "perla",
    tone_labels: List[str] = None,
    convert_to_black_white: bool = False,
    n_dominant_colors=2,
    new_width=250,
    scale=1.1,
    min_nbrs=5,
    min_size=(90, 90),
    threshold=0.15,
    return_report_image=False,
):
    """
    Process the image and return the result.
    :param filename_or_url: The filename (in local devices) or URL (in Internet) of the image.
    :param image_type: Specify whether the input image(s) is/are colored or black/white.
           Valid choices are: "auto", "color" or "bw", Defaults to "auto", which will be detected automatically.
    :param tone_palette: Skin tone palette; Valid choices can be `perla`, `yadon-ostfeld`, `proder`;
           You can also input RGB hex value leading by "#" or RGB values separated by comma(,).
           E.g., ['#373028', '#422811'] or ['255,255,255', '100,100,100']
    :param tone_labels: Skin tone labels; default values are the uppercase alphabet list leading by the image type
           ('C' for 'color'; 'B' for 'Black&White'), e.g., ['CA', 'CB', ..., 'CZ'] or ['BA', 'BB', ..., 'BZ'].
    :param convert_to_black_white: Whether to convert the image to black/white before processing. Defaults to False.
    :param n_dominant_colors: Number of dominant colors to be extracted from the image. Defaults to 2.
    :param new_width: Resize the images with the specified width. Negative value will be ignored, defaults to 250.
    :param scale: How much the image size is reduced at each image scale. Defaults to 1.1.
    :param min_nbrs: How many neighbors each candidate rectangle should have to retain it.
           Higher value results in less detection but with higher quality, defaults to 5.
    :param min_size: Minimum possible face size. Faces smaller than that are ignored, defaults to (90, 90).
    :param threshold: What percentage of the skin area is required to identify the face, defaults to 0.15.
    :param return_report_image: Whether to return the report image(s) in the result. Defaults to False.
    :return:
    """
    image, basename, extension = load_image(filename_or_url, flags=cv2.IMREAD_COLOR)
    if image is None:
        msg = f"{basename}{extension} is not found or is not a valid image."
        LOG.error(msg)
        return {
            "filename": basename,
            "message": msg,
        }

    is_bw = is_black_white(image)
    decoded_image_type = image_type
    if image_type == "auto":
        decoded_image_type = "bw" if convert_to_black_white or is_bw else "color"
    else:
        is_bw = image_type == "bw"

    if not tone_palette:
        tone_palette = "bw" if decoded_image_type == "bw" else "perla"
    if len(tone_palette) == 1:
        tone_palette = tone_palette[0]

    if isinstance(tone_palette, str):
        tone_palette = tone_palette.lower()
        if tone_palette not in DEFAULT_TONE_PALETTE:
            raise ArgumentError(f"Invalid `tone_palette`: {tone_palette}, valid choices are: {DEFAULT_TONE_PALETTE.keys()}")
        skin_tone_palette = DEFAULT_TONE_PALETTE[tone_palette]
    else:
        skin_tone_palette = normalize_palette(tone_palette)

    skin_tone_labels = tone_labels or default_tone_labels(skin_tone_palette, "C" if decoded_image_type == "color" else "B")
    if len(skin_tone_palette) != len(skin_tone_labels):
        raise ArgumentError("Argument -p/--palette and -l/--labels must have the same length.")

    records, report_images = process_image(
        image,
        is_bw,
        convert_to_black_white,
        skin_tone_palette,
        skin_tone_labels,
        new_width=new_width,
        n_dominant_colors=n_dominant_colors,
        scaleFactor=scale,
        minNeighbors=min_nbrs,
        minSize=min_size,
        threshold=threshold,
        verbose=return_report_image,
    )
    return {
        "basename": basename,
        "extension": extension,
        "image_type": decoded_image_type,
        "faces": records,
        "report_images": report_images,
    }



import functools
import logging
import os
import shutil
import sys
import threading
from datetime import datetime
from multiprocessing import freeze_support, cpu_count, Pool

import cv2
import numpy as np
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm
from typing import List

from stone.api import process
from stone.image import normalize_palette
from stone.utils import (
    build_arguments,
    build_image_paths,
    is_windows,
    ArgumentError,
    is_debugging,
    resolve_labels,
)
from stone.package import (
    __app_name__,
    __version__,
    __description__,
    __copyright__,
    __url__,
    __author__,
    __license__,
    __code__,
    __issues__,
    __package_name__,
)

LOG = logging.getLogger(__name__)
lock = threading.Lock()

use_cli = len(sys.argv) > 1 and "--gui" not in sys.argv


def process_in_main(
    filename_or_url,
    image_type,
    tone_palette,
    tone_labels,
    convert_to_black_white,
    n_dominant_colors=2,
    new_width=250,
    scale=1.1,
    min_nbrs=5,
    min_size=(90, 90),
    threshold=0.3,
    return_report_image=False,
):
    """
    This is a wrapper function that calls process() in the main process to avoid pickling error.
    :param filename_or_url:
    :param image_type:
    :param tone_palette:
    :param tone_labels:
    :param convert_to_black_white:
    :param n_dominant_colors:
    :param new_width:
    :param scale:
    :param min_nbrs:
    :param min_size:
    :param threshold:
    :param return_report_image:
    :return:
    """
    try:
        return process(
            filename_or_url,
            image_type=image_type,
            tone_palette=tone_palette,
            tone_labels=tone_labels,
            convert_to_black_white=convert_to_black_white,
            n_dominant_colors=n_dominant_colors,
            new_width=new_width,
            scale=scale,
            min_nbrs=min_nbrs,
            min_size=min_size,
            threshold=threshold,
            return_report_image=return_report_image,
        )
    except ArgumentError as e:
        # Abort the app if any argument error occurs
        raise e
    except Exception as e:
        msg = f"Error processing image {filename_or_url}: {str(e)}"
        LOG.error(msg)
        return {
            "filename": filename_or_url,
            "message": msg,
        }


def main():
    args = build_arguments()
    # Setup logger
    now = datetime.now()
    output_dir = args.output
    log_dir = os.path.join(output_dir, "./log")
    os.makedirs(log_dir, exist_ok=True)

    logging.basicConfig(
        filename=now.strftime(f"{log_dir}/log-%y%m%d%H%M.log"),
        level=logging.INFO,
        format="[%(asctime)s] {%(filename)s:%(lineno)4d} %(levelname)s - %(message)s",
        datefmt="%H:%M:%S",
    )

    image_paths = build_image_paths(args.images, args.recursive)

    debug: bool = args.debug
    to_bw: bool = args.black_white

    specified_palette: List[str] = args.palette

    specified_tone_labels = resolve_labels(args.labels)

    new_width = args.new_width
    n_dominant_colors = args.n_colors
    min_size = args.min_size[:2]
    scale = args.scale
    min_nbrs = args.min_nbrs

    os.makedirs(output_dir, exist_ok=True)
    result_filename = os.path.join(output_dir, "./result.csv")
    image_type_setting = args.image_type
    threshold = args.threshold

    def write_to_csv(row: list):
        with lock:
            with open(result_filename, "a", newline="", encoding="UTF8") as f:
                f.write(",".join(map(str, row)) + "\n")

    num_workers = cpu_count() if args.n_workers == 0 else args.n_workers

    pool = Pool(processes=num_workers)

    # Backup result.csv if exists
    if os.path.exists(result_filename):
        renamed_file = os.path.join(output_dir, now.strftime("./result_bak_%y%m%d%H%M.csv"))
        shutil.move(result_filename, renamed_file)
    header = (
        "file,image type,face id,"
        + ",".join([f"dominant {i + 1},percent {i + 1}" for i in range(n_dominant_colors)])
        + ",skin tone,tone label,accuracy(0-100)"
    )
    write_to_csv(header.split(","))

    # Start
    process_wrapper = functools.partial(
        process if is_debugging() else process_in_main,
        image_type=image_type_setting,
        tone_palette=specified_palette,
        tone_labels=specified_tone_labels,
        convert_to_black_white=to_bw,
        n_dominant_colors=n_dominant_colors,
        new_width=new_width,
        scale=scale,
        min_nbrs=min_nbrs,
        min_size=min_size,
        threshold=threshold,
        return_report_image=debug,
    )
    print("The program is processing your images...")
    print("Please wait for the program to finish.")
    with logging_redirect_tqdm():
        with tqdm(image_paths, desc="Processing images", unit="images") as pbar:
            for result in pool.imap(process_wrapper, image_paths):
                if "message" in result:
                    write_to_csv([result["filename"], result["message"]])
                    pbar.update()
                    continue

                basename = result["basename"]
                extension = result["extension"]
                image_type = result["image_type"]
                faces = result["faces"]
                report_images = result["report_images"]

                pbar.set_description(f"Processing {basename}")
                n_faces = len(faces)

                for face_record in faces:
                    face_id = face_record["face_id"]
                    if face_id == "NA":
                        n_faces = 0  # Did not detect any faces
                    dominant_colors = [[item["color"], item["percent"]] for item in face_record["dominant_colors"]]
                    record = (
                        [f"{basename}{extension}", image_type, face_id]
                        + np.hstack(dominant_colors).tolist()
                        + [face_record["skin_tone"], face_record["tone_label"], face_record["accuracy"]]
                    )

                    write_to_csv(record)
                    pbar.set_postfix(
                        {
                            "Image Type": image_type,
                            "#Faces": n_faces,
                            "Face ID": face_id,
                            "Skin Tone": face_record["skin_tone"],
                            "Label": face_record["tone_label"],
                            "Accuracy": face_record["accuracy"],
                        }
                    )
                if debug:
                    debug_dir = os.path.join(output_dir, f"./debug/{image_type}/faces_{n_faces}")
                    os.makedirs(debug_dir, exist_ok=True)
                    for face_id, report_image in report_images.items():
                        image_name = f"{basename}-{face_id}"
                        report_filename = os.path.join(debug_dir, f"{image_name}{extension}")
                        with lock:
                            cv2.imwrite(report_filename, report_image)
                pbar.update()
    pool.close()
    pool.join()


sys.argv.remove("--gui") if "--gui" in sys.argv else None
if not use_cli and "--ignore-gooey" not in sys.argv:
    try:
        from gooey import Gooey
    except ImportError:
        # If gooey is not installed, use a dummy decorator
        from stone.utils import Gooey

    from importlib.resources import files

    main = Gooey(
        show_preview_warning=False,
        advanced=True,  # fixme: `False` is not working
        dump_build_config=False,  # fixme: `True` is not working, as the path cannot be resolved correctly
        target="stone",
        suppress_gooey_flag=True,
        program_name=f"{__app_name__} v{__version__}",
        required_cols=1,
        optional_cols=1,
        image_dir=str(files("stone.ui")),
        tabbed_groups=True,
        navigation="Tabbed",
        richtext_controls=True,
        use_cmd_args=True,
        menu=[
            {
                "name": "Help",
                "items": [
                    {
                        "type": "AboutDialog",
                        "menuTitle": "About",
                        "name": __app_name__,
                        "description": __description__,
                        "version": __version__,
                        "copyright": __copyright__,
                        "website": __url__,
                        "developer": __author__,
                        "license": __license__,
                    },
                    {"type": "Link", "menuTitle": "Documentation", "url": __code__},
                    {"type": "Link", "menuTitle": "Report Bugs", "url": __issues__},
                ],
            },
        ],
    )(main)

if __name__ == "__main__":
    if is_windows():
        freeze_support()
    main()
